@@ -1,5 +1,5 @@
 /* server.c --- DIGEST-MD5 mechanism from RFC 2831, server side.
- * Copyright (C) 2002, 2003, 2004, 2006  Simon Josefsson
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009  Simon Josefsson
  *
  * This file is part of GNU SASL Library.
  *
@@ -20,7 +20,7 @@
  *
  */
 
-#if HAVE_CONFIG_H
+#ifdef HAVE_CONFIG_H
 # include "config.h"
 #endif
 
@@ -34,6 +34,7 @@
 #include <string.h>
 
 /* Get tools. */
+#include "nonascii.h"
 #include "tokens.h"
 #include "parser.h"
 #include "printer.h"
@@ -41,6 +42,7 @@
 #include "session.h"
 #include "digesthmac.h"
 #include "validate.h"
+#include "qop.h"
 
 #define NONCE_ENTROPY_BYTES 16
 
@@ -82,7 +84,7 @@ _gsasl_digest_md5_server_start (Gsasl_session * sctx, void **mech_data)
       return GSASL_MALLOC_ERROR;
     }
 
-  state->challenge.qops = DIGEST_MD5_QOP_AUTH | DIGEST_MD5_QOP_AUTH_INT;
+  state->challenge.qops = DIGEST_MD5_QOP_AUTH;
   state->challenge.ciphers = 0;
 
   state->challenge.nonce = p;
@@ -93,6 +95,50 @@ _gsasl_digest_md5_server_start (Gsasl_session * sctx, void **mech_data)
   return GSASL_OK;
 }
 
+static char
+_gsasl_digest_md5_hexdigit_to_char (char hexdigit)
+{
+  /* The hex representation always contains lowercase alphabetic
+     characters.  See RFC 2831, 1.1. */
+
+  if (hexdigit >= '0' && hexdigit <= '9')
+    return hexdigit - '0';
+  if (hexdigit >= 'a' && hexdigit <= 'z')
+    return hexdigit - 'a' + 10;
+
+  return -1;
+}
+
+static char
+_gsasl_digest_md5_hex_to_char (char u, char l)
+{
+  return (char) (((unsigned char) _gsasl_digest_md5_hexdigit_to_char (u)) *
+		 16 + _gsasl_digest_md5_hexdigit_to_char (l));
+}
+
+static int
+_gsasl_digest_md5_set_hashed_secret (char *secret, const char *hex_secret)
+{
+  /* Convert the hex string containing the secret to a byte array */
+  const char *p;
+  char *s;
+
+  if (!hex_secret)
+    return GSASL_AUTHENTICATION_ERROR;
+
+  s = secret;
+  p = hex_secret;
+  while (*p)
+    {
+      *s = _gsasl_digest_md5_hex_to_char (p[0], p[1]);
+      s++;
+
+      p += 2;
+    }
+
+  return GSASL_OK;
+}
+
 int
 _gsasl_digest_md5_server_step (Gsasl_session * sctx,
 			       void *mech_data,
@@ -128,7 +174,27 @@ _gsasl_digest_md5_server_step (Gsasl_session * sctx,
 	  }
       }
 
-      /* FIXME: qop, cipher, maxbuf, more realms. */
+      /* Set QOP */
+      {
+	const char *qopstr = gsasl_property_get (sctx, GSASL_QOPS);
+
+	if (qopstr)
+	  {
+	    int qops = digest_md5_qopstr2qops (qopstr);
+
+	    if (qops == -1)
+	      return GSASL_MALLOC_ERROR;
+
+	    /* We don't support confidentiality right now. */
+	    if (qops & DIGEST_MD5_QOP_AUTH_CONF)
+	      return GSASL_AUTHENTICATION_ERROR;
+
+	    if (qops)
+	      state->challenge.qops = qops;
+	  }
+      }
+
+      /* FIXME: cipher, maxbuf, more realms. */
 
       /* Create challenge. */
       *output = digest_md5_print_challenge (&state->challenge);
@@ -149,33 +215,75 @@ _gsasl_digest_md5_server_step (Gsasl_session * sctx,
 	return GSASL_MECHANISM_PARSE_ERROR;
 
       /* Store properties, from the client response. */
-      gsasl_property_set (sctx, GSASL_AUTHID, state->response.username);
+      if (state->response.utf8)
+	{
+	  gsasl_property_set (sctx, GSASL_AUTHID, state->response.username);
+	  gsasl_property_set (sctx, GSASL_REALM, state->response.realm);
+	}
+      else
+	{
+	  /* Client provided username/realm in ISO-8859-1 form,
+	     convert it to UTF-8 since the library is all-UTF-8. */
+	  char *tmp;
+
+	  tmp = latin1toutf8 (state->response.username);
+	  if (!tmp)
+	    return GSASL_MALLOC_ERROR;
+	  gsasl_property_set (sctx, GSASL_AUTHID, tmp);
+	  free (tmp);
+
+	  tmp = latin1toutf8 (state->response.realm);
+	  if (!tmp)
+	    return GSASL_MALLOC_ERROR;
+	  gsasl_property_set (sctx, GSASL_REALM, tmp);
+	  free (tmp);
+	}
       gsasl_property_set (sctx, GSASL_AUTHZID, state->response.authzid);
-      gsasl_property_set (sctx, GSASL_REALM, state->response.realm);
 
-      /* FIXME: qop, cipher, maxbuf. */
+      /* FIXME: cipher, maxbuf.  */
 
-      /* Compute secret.  TODO: Add callback to retrieve hashed
-         secret. */
+      /* Compute secret. */
       {
-	const char *c;
-	char *tmp, *tmp2;
+	const char *passwd;
+	const char *hashed_passwd;
 
-	c = gsasl_property_get (sctx, GSASL_PASSWORD);
-	if (!c)
-	  return GSASL_NO_PASSWORD;
+	hashed_passwd =
+	  gsasl_property_get (sctx, GSASL_DIGEST_MD5_HASHED_PASSWORD);
+	if (hashed_passwd)
+	  {
+	    if (strlen (hashed_passwd) != (DIGEST_MD5_LENGTH * 2))
+	      return GSASL_AUTHENTICATION_ERROR;
 
-	if (asprintf (&tmp, "%s:%s:%s", state->response.username,
-		      state->response.realm ?
-		      state->response.realm : "", c) < 0)
-	  return GSASL_MALLOC_ERROR;
+	    rc = _gsasl_digest_md5_set_hashed_secret (state->secret,
+						      hashed_passwd);
+	    if (rc != GSASL_OK)
+	      return rc;
+	  }
+	else if ((passwd = gsasl_property_get (sctx, GSASL_PASSWORD)) != NULL)
+	  {
+	    char *tmp, *tmp2;
 
-	rc = gsasl_md5 (tmp, strlen (tmp), &tmp2);
-	free (tmp);
-	if (rc != GSASL_OK)
-	  return rc;
-	memcpy (state->secret, tmp2, DIGEST_MD5_LENGTH);
-	free (tmp2);
+	    tmp2 = utf8tolatin1ifpossible (passwd);
+
+	    rc = asprintf (&tmp, "%s:%s:%s", state->response.username,
+			   state->response.realm ?
+			   state->response.realm : "", tmp2);
+	    free (tmp2);
+	    if (rc < 0)
+	      return GSASL_MALLOC_ERROR;
+
+	    rc = gsasl_md5 (tmp, strlen (tmp), &tmp2);
+	    free (tmp);
+	    if (rc != GSASL_OK)
+	      return rc;
+
+	    memcpy (state->secret, tmp2, DIGEST_MD5_LENGTH);
+	    free (tmp2);
+	  }
+	else
+	  {
+	    return GSASL_NO_PASSWORD;
+	  }
       }
 
       /* Check client response. */
@@ -187,7 +295,8 @@ _gsasl_digest_md5_server_step (Gsasl_session * sctx,
 			      state->response.cnonce, state->response.qop,
 			      state->response.authzid,
 			      state->response.digesturi, 0,
-			      state->response.cipher, NULL, NULL, NULL, NULL);
+			      state->response.cipher,
+			      state->kic, state->kis, state->kcc, state->kcs);
 	if (rc)
 	  return GSASL_AUTHENTICATION_ERROR;
 
