@@ -1,5 +1,5 @@
 /* obsolete.c --- Obsolete functions kept around for backwards compatibility.
- * Copyright (C) 2002, 2003, 2004, 2005, 2006  Simon Josefsson
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009  Simon Josefsson
  *
  * This file is part of GNU SASL Library.
  *
@@ -22,6 +22,10 @@
 
 #include "internal.h"
 
+#if USE_DIGEST_MD5
+# include "qop.h"
+#endif
+
 /**
  * gsasl_client_listmech:
  * @ctx: libgsasl handle.
@@ -31,9 +35,10 @@
  *
  * Write SASL names, separated by space, of mechanisms supported by
  * the libgsasl client to the output array.  To find out how large the
- * output array must be, call this function with out=NULL.
+ * output array must be, call this function with a %NULL @out
+ * parameter.
  *
- * Return value: Returns GSASL_OK if successful, or error code.
+ * Return value: Returns %GSASL_OK if successful, or error code.
  *
  * Deprecated: Use gsasl_client_mechlist() instead.
  **/
@@ -73,9 +78,10 @@ gsasl_client_listmech (Gsasl * ctx, char *out, size_t * outlen)
  *
  * Write SASL names, separated by space, of mechanisms supported by
  * the libgsasl server to the output array.  To find out how large the
- * output array must be, call this function with out=NULL.
+ * output array must be, call this function with a %NULL @out
+ * parameter.
  *
- * Return value: Returns GSASL_OK if successful, or error code.
+ * Return value: Returns %GSASL_OK if successful, or error code.
  *
  * Deprecated: Use gsasl_server_mechlist() instead.
  **/
@@ -149,11 +155,11 @@ _gsasl_step (Gsasl_session * sctx,
  * data to server (into variables output and output_len).
  *
  * The contents of the output buffer is unspecified if this functions
- * returns anything other than GSASL_NEEDS_MORE.
+ * returns anything other than %GSASL_NEEDS_MORE.
  *
- * Return value: Returns GSASL_OK if authenticated terminated
- * successfully, GSASL_NEEDS_MORE if more data is needed, or error
- * code.
+ * Return value: Returns %GSASL_OK if authenticated terminated
+ *   successfully, %GSASL_NEEDS_MORE if more data is needed, or error
+ *   code.
  *
  * Deprecated: Use gsasl_step() instead.
  **/
@@ -179,11 +185,11 @@ gsasl_client_step (Gsasl_session * sctx,
  * data to client (into variables output and output_len).
  *
  * The contents of the output buffer is unspecified if this functions
- * returns anything other than GSASL_NEEDS_MORE.
+ * returns anything other than %GSASL_NEEDS_MORE.
  *
- * Return value: Returns GSASL_OK if authenticated terminated
- * successfully, GSASL_NEEDS_MORE if more data is needed, or error
- * code.
+ * Return value: Returns %GSASL_OK if authenticated terminated
+ *   successfully, %GSASL_NEEDS_MORE if more data is needed, or error
+ *   code.
  *
  * Deprecated: Use gsasl_step() instead.
  **/
@@ -339,7 +345,7 @@ gsasl_client_application_data_set (Gsasl_session * sctx,
  * It is normally used by the application to maintain state between
  * the main program and the callback.
  *
- * Return value: Returns the application specific data, or NULL.
+ * Return value: Returns the application specific data, or %NULL.
  *
  * Deprecated: Use gsasl_callback_hook_get() or
  * gsasl_session_hook_get() instead.
@@ -394,7 +400,7 @@ gsasl_server_application_data_set (Gsasl_session * sctx,
  * It is normally used by the application to maintain state between
  * the main program and the callback.
  *
- * Return value: Returns the application specific data, or NULL.
+ * Return value: Returns the application specific data, or %NULL.
  *
  * Deprecated: Use gsasl_callback_hook_get() or
  * gsasl_session_hook_get() instead.
@@ -451,8 +457,8 @@ gsasl_ctx_get (Gsasl_session * sctx)
  * Encode data according to negotiated SASL mechanism.  This might mean
  * that data is integrity or privacy protected.
  *
- * Return value: Returns GSASL_OK if encoding was successful, otherwise
- * an error code.
+ * Return value: Returns %GSASL_OK if encoding was successful,
+ *   otherwise an error code.
  *
  * Deprecated: Use gsasl_encode() instead.
  *
@@ -491,8 +497,8 @@ gsasl_encode_inline (Gsasl_session * sctx,
  * Decode data according to negotiated SASL mechanism.  This might mean
  * that data is integrity or privacy protected.
  *
- * Return value: Returns GSASL_OK if encoding was successful, otherwise
- * an error code.
+ * Return value: Returns %GSASL_OK if encoding was successful,
+ *   otherwise an error code.
  *
  * Deprecated: Use gsasl_decode() instead.
  *
@@ -548,7 +554,7 @@ gsasl_application_data_set (Gsasl * ctx, void *appdata)
  * normally used by the application to maintain state between the main
  * program and the callback.
  *
- * Return value: Returns the application specific data, or NULL.
+ * Return value: Returns the application specific data, or %NULL.
  *
  * Deprecated: Use gsasl_callback_hook_get() instead.
  **/
@@ -586,7 +592,7 @@ gsasl_appinfo_set (Gsasl_session * sctx, void *appdata)
  * is normally used by the application to maintain state between the
  * main program and the callback.
  *
- * Return value: Returns the application specific data, or NULL.
+ * Return value: Returns the application specific data, or %NULL.
  *
  * Deprecated: Use gsasl_callback_hook_get() instead.
  **/
@@ -1545,7 +1551,7 @@ gsasl_server_callback_service_get (Gsasl * ctx)
   return ctx ? ctx->cbs_service : NULL;
 }
 
-#if WITH_SASLPREP
+#if HAVE_LIBIDN
 # include <stringprep.h>
 #endif
 
@@ -1580,7 +1586,7 @@ gsasl_stringprep_nfkc (const char *in, ssize_t len)
 {
   char *out = NULL;
 
-#if WITH_SASLPREP
+#if HAVE_LIBIDN
   out = stringprep_utf8_nfkc_normalize (in, len);
 #endif
 
@@ -1610,7 +1616,7 @@ char *
 gsasl_stringprep_saslprep (const char *in, int *stringprep_rc)
 {
   char *out = NULL;
-#if WITH_SASLPREP
+#if HAVE_LIBIDN
   int rc;
 
   rc = stringprep_profile (in, &out, "SASLprep", 0);
@@ -1644,7 +1650,7 @@ char *
 gsasl_stringprep_trace (const char *in, int *stringprep_rc)
 {
   char *out = NULL;
-#if WITH_SASLPREP
+#if HAVE_LIBIDN
   int rc;
 
   rc = stringprep_profile (in, &out, "trace", 0);
@@ -1685,107 +1691,39 @@ int
 gsasl_md5pwd_get_password (const char *filename,
 			   const char *username, char *key, size_t * keylen)
 {
-  char matchbuf[BUFSIZ];
-  char line[BUFSIZ];
+  char *tmp;
+  size_t tmplen;
+  int res;
   FILE *fh;
 
   fh = fopen (filename, "r");
   if (fh == NULL)
     return GSASL_FOPEN_ERROR;
+  fclose (fh);
 
-  sprintf (matchbuf, "%s\t", username);
+  res = gsasl_simple_getpass (filename, username, &tmp);
+  if (res != GSASL_OK)
+    return res;
 
-  while (!feof (fh))
+  tmplen = strlen (tmp);
+
+  if (*keylen < tmplen + 1)
     {
-      if (fgets (line, BUFSIZ, fh) == NULL)
-	break;
-
-      if (line[0] == '#')
-	continue;
-
-      while (strlen (line) > 0 && (line[strlen (line) - 1] == '\n' ||
-				   line[strlen (line) - 1] == '\r'))
-	line[strlen (line) - 1] = '\0';
-
-      if (strlen (line) <= strlen (matchbuf))
-	continue;
-
-      if (strncmp (matchbuf, line, strlen (matchbuf)) == 0)
-	{
-	  if (*keylen < strlen (line) - strlen (matchbuf))
-	    {
-	      fclose (fh);
-	      return GSASL_TOO_SMALL_BUFFER;
-	    }
-
-	  *keylen = strlen (line) - strlen (matchbuf);
-
-	  if (key)
-	    memcpy (key, &line[strlen (matchbuf)], *keylen);
-
-	  fclose (fh);
-
-	  return GSASL_OK;
-	}
+      free (tmp);
+      return GSASL_TOO_SMALL_BUFFER;
     }
 
-  if (fclose (fh) != 0)
-    return GSASL_FCLOSE_ERROR;
+  *keylen = tmplen;
 
-  return GSASL_AUTHENTICATION_ERROR;
+  if (key)
+    memcpy (key, tmp, tmplen);
+
+  free (tmp);
+
+  return GSASL_OK;
 }
 
-
-/*
- * Copyright (c) 1996-1999 by Internet Software Consortium.
- *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
- *
- * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM DISCLAIMS
- * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET SOFTWARE
- * CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
- * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
- * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
- * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
- * SOFTWARE.
- */
-
-/*
- * Portions Copyright (c) 1995 by International Business Machines, Inc.
- *
- * International Business Machines, Inc. (hereinafter called IBM) grants
- * permission under its copyrights to use, copy, modify, and distribute this
- * Software with or without fee, provided that the above copyright notice and
- * all paragraphs of this notice appear in all copies, and that the name of IBM
- * not be used in connection with the marketing of any product incorporating
- * the Software or modifications thereof, without specific, written prior
- * permission.
- *
- * To the extent it has a right to do so, IBM grants an immunity from suit
- * under its patents, if any, for the use, sale or manufacture of products to
- * the extent that such products are used for performing Domain Name System
- * dynamic updates in TCP/IP networks by means of the Software.  No immunity is
- * granted for any product per se or for any other function of any product.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", AND IBM DISCLAIMS ALL WARRANTIES,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
- * PARTICULAR PURPOSE.  IN NO EVENT SHALL IBM BE LIABLE FOR ANY SPECIAL,
- * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER ARISING
- * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE, EVEN
- * IF IBM IS APPRISED OF THE POSSIBILITY OF SUCH DAMAGES.
- */
-
-#include <ctype.h>
-#include <stdio.h>
-
-#define Assert(Cond) if (!(Cond)) abort()
-
-static const char Base64[] =
-  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
-static const char Pad64 = '=';
+#include <minmax.h>
 
 /**
  * gsasl_base64_encode:
@@ -1807,64 +1745,20 @@ int
 gsasl_base64_encode (char const *src,
 		     size_t srclength, char *target, size_t targsize)
 {
-  size_t datalength = 0;
-  unsigned char input[3];
-  unsigned char output[4];
-  size_t i;
+  int rc;
+  char *out;
+  size_t outlen;
+  int copied;
 
-  while (2 < srclength)
-    {
-      input[0] = *src++;
-      input[1] = *src++;
-      input[2] = *src++;
-      srclength -= 3;
+  rc = gsasl_base64_to (src, srclength, &out, &outlen);
+  if (rc)
+    return -1;
 
-      output[0] = input[0] >> 2;
-      output[1] = ((input[0] & 0x03) << 4) + (input[1] >> 4);
-      output[2] = ((input[1] & 0x0f) << 2) + (input[2] >> 6);
-      output[3] = input[2] & 0x3f;
-      Assert (output[0] < 64);
-      Assert (output[1] < 64);
-      Assert (output[2] < 64);
-      Assert (output[3] < 64);
+  copied = MIN (outlen, targsize);
+  memcpy (target, out, copied);
+  free (out);
 
-      if (datalength + 4 > targsize)
-	return (-1);
-      target[datalength++] = Base64[output[0]];
-      target[datalength++] = Base64[output[1]];
-      target[datalength++] = Base64[output[2]];
-      target[datalength++] = Base64[output[3]];
-    }
-
-  /* Now we worry about padding. */
-  if (0 != srclength)
-    {
-      /* Get what's left. */
-      input[0] = input[1] = input[2] = '\0';
-      for (i = 0; i < srclength; i++)
-	input[i] = *src++;
-
-      output[0] = input[0] >> 2;
-      output[1] = ((input[0] & 0x03) << 4) + (input[1] >> 4);
-      output[2] = ((input[1] & 0x0f) << 2) + (input[2] >> 6);
-      Assert (output[0] < 64);
-      Assert (output[1] < 64);
-      Assert (output[2] < 64);
-
-      if (datalength + 4 > targsize)
-	return (-1);
-      target[datalength++] = Base64[output[0]];
-      target[datalength++] = Base64[output[1]];
-      if (srclength == 1)
-	target[datalength++] = Pad64;
-      else
-	target[datalength++] = Base64[output[2]];
-      target[datalength++] = Pad64;
-    }
-  if (datalength >= targsize)
-    return (-1);
-  target[datalength] = '\0';	/* Returned value doesn't count \0. */
-  return (datalength);
+  return copied;
 }
 
 /**
@@ -1885,126 +1779,319 @@ gsasl_base64_encode (char const *src,
 int
 gsasl_base64_decode (char const *src, char *target, size_t targsize)
 {
-  int tarindex, state, ch;
-  char *pos;
+  int rc;
+  char *out;
+  size_t outlen;
+  int copied;
 
-  state = 0;
-  tarindex = 0;
+  rc = gsasl_base64_from (src, strlen (src), &out, &outlen);
+  if (rc)
+    return -1;
 
-  while ((ch = *src++) != '\0')
-    {
-      if (isspace (ch))		/* Skip whitespace anywhere. */
-	continue;
+  copied = MIN (outlen, targsize);
+  memcpy (target, out, copied);
+  free (out);
 
-      if (ch == Pad64)
-	break;
-
-      pos = strchr (Base64, ch);
-      if (pos == 0)		/* A non-base64 character. */
-	return (-1);
-
-      switch (state)
-	{
-	case 0:
-	  if (target)
-	    {
-	      if ((size_t) tarindex >= targsize)
-		return (-1);
-	      target[tarindex] = (pos - Base64) << 2;
-	    }
-	  state = 1;
-	  break;
-	case 1:
-	  if (target)
-	    {
-	      if ((size_t) tarindex + 1 >= targsize)
-		return (-1);
-	      target[tarindex] |= (pos - Base64) >> 4;
-	      target[tarindex + 1] = ((pos - Base64) & 0x0f) << 4;
-	    }
-	  tarindex++;
-	  state = 2;
-	  break;
-	case 2:
-	  if (target)
-	    {
-	      if ((size_t) tarindex + 1 >= targsize)
-		return (-1);
-	      target[tarindex] |= (pos - Base64) >> 2;
-	      target[tarindex + 1] = ((pos - Base64) & 0x03) << 6;
-	    }
-	  tarindex++;
-	  state = 3;
-	  break;
-	case 3:
-	  if (target)
-	    {
-	      if ((size_t) tarindex >= targsize)
-		return (-1);
-	      target[tarindex] |= (pos - Base64);
-	    }
-	  tarindex++;
-	  state = 0;
-	  break;
-	default:
-	  abort ();
-	}
+  return copied;
+}
+
+static const char *
+pmap (Gsasl_session * sctx, Gsasl_property prop, char *buf, size_t buflen)
+{
+  int res;
+
+  buf[0] = '\0';
+
+  /* Translate obsolete callbacks to modern properties. */
+
+  switch (prop)
+    {
+    case GSASL_SERVICE:
+      {
+	Gsasl_client_callback_service cb_service
+	  = gsasl_client_callback_service_get (sctx->ctx);
+	if (!cb_service)
+	  break;
+	res = cb_service (sctx, buf, &buflen, NULL, 0, NULL, 0);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_HOSTNAME:
+      {
+	Gsasl_client_callback_service cb_service
+	  = gsasl_client_callback_service_get (sctx->ctx);
+	if (!cb_service)
+	  break;
+	res = cb_service (sctx, NULL, 0, buf, &buflen, NULL, 0);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_ANONYMOUS_TOKEN:
+      {
+	Gsasl_client_callback_anonymous cb_anonymous
+	  = gsasl_client_callback_anonymous_get (sctx->ctx);
+	if (!cb_anonymous)
+	  break;
+	res = cb_anonymous (sctx, buf, &buflen);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_AUTHID:
+      {
+	Gsasl_client_callback_authentication_id cb_authentication_id
+	  = gsasl_client_callback_authentication_id_get (sctx->ctx);
+	if (!cb_authentication_id)
+	  break;
+	res = cb_authentication_id (sctx, buf, &buflen);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_AUTHZID:
+      {
+	Gsasl_client_callback_authorization_id cb_authorization_id
+	  = gsasl_client_callback_authorization_id_get (sctx->ctx);
+	if (!cb_authorization_id)
+	  break;
+	res = cb_authorization_id (sctx, buf, &buflen);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_PASSWORD:
+      {
+	Gsasl_client_callback_password cb_password
+	  = gsasl_client_callback_password_get (sctx->ctx);
+	if (!cb_password)
+	  break;
+	res = cb_password (sctx, buf, &buflen);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_PASSCODE:
+      {
+	Gsasl_client_callback_passcode cb_passcode
+	  = gsasl_client_callback_passcode_get (sctx->ctx);
+	if (!cb_passcode)
+	  break;
+	res = cb_passcode (sctx, buf, &buflen);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_PIN:
+      {
+	Gsasl_client_callback_pin cb_pin
+	  = gsasl_client_callback_pin_get (sctx->ctx);
+	if (!cb_pin)
+	  break;
+	res = cb_pin (sctx, sctx->suggestedpin, buf, &buflen);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+    case GSASL_REALM:
+      {
+	Gsasl_client_callback_realm cb_realm
+	  = gsasl_client_callback_realm_get (sctx->ctx);
+	if (!cb_realm)
+	  break;
+	res = cb_realm (sctx, buf, &buflen);
+	if (res != GSASL_OK)
+	  break;
+	buf[buflen] = '\0';
+	gsasl_property_set (sctx, prop, buf);
+	break;
+      }
+
+#if USE_DIGEST_MD5
+    case GSASL_QOP:
+      {
+	Gsasl_client_callback_qop cb_qop
+	  = gsasl_client_callback_qop_get (sctx->ctx);
+	int serverqops;
+	Gsasl_qop qop;
+	if (!cb_qop)
+	  break;
+	serverqops = digest_md5_qopstr2qops (sctx->qops);
+	if (serverqops == -1)
+	  return NULL;
+	qop = cb_qop (sctx, serverqops);
+	if (qop & 0x07)
+	  gsasl_property_set (sctx, prop, digest_md5_qops2qopstr (qop));
+	break;
+      }
+      break;
+#endif
+
+    default:
+      break;
     }
+
+  return gsasl_property_fast (sctx, prop);
+}
+
+const char *
+_gsasl_obsolete_property_map (Gsasl_session * sctx, Gsasl_property prop)
+{
+  const char *ret;
+  char *buf;
+
+  buf = malloc (BUFSIZ);
+  if (!buf)
+    return NULL;
+
+  ret = pmap (sctx, prop, buf, BUFSIZ - 1);
 
-  /*
-   * We are done decoding Base-64 chars.  Let's see if we ended
-   * on a byte boundary, and/or with erroneous trailing characters.
-   */
-
-  if (ch == Pad64)
-    {				/* We got a pad char. */
-      ch = *src++;		/* Skip it, get next. */
-      switch (state)
-	{
-	case 0:		/* Invalid = in first position */
-	case 1:		/* Invalid = in second position */
-	  return (-1);
-
-	case 2:		/* Valid, means one byte of info */
-	  /* Skip any number of spaces. */
-	  for ((void) NULL; ch != '\0'; ch = *src++)
-	    if (!isspace (ch))
-	      break;
-	  /* Make sure there is another trailing = sign. */
-	  if (ch != Pad64)
-	    return (-1);
-	  ch = *src++;		/* Skip the = */
-	  /* Fall through to "single trailing =" case. */
-	  /* FALLTHROUGH */
-
-	case 3:		/* Valid, means two bytes of info */
-	  /*
-	   * We know this char is an =.  Is there anything but
-	   * whitespace after it?
-	   */
-	  for ((void) NULL; ch != '\0'; ch = *src++)
-	    if (!isspace (ch))
-	      return (-1);
-
-	  /*
-	   * Now make sure for cases 2 and 3 that the "extra"
-	   * bits that slopped past the last full byte were
-	   * zeros.  If we don't check them, they become a
-	   * subliminal channel.
-	   */
-	  if (target && target[tarindex] != 0)
-	    return (-1);
-	}
-    }
-  else
+  free (buf);
+
+  return ret;
+}
+
+int
+_gsasl_obsolete_callback (Gsasl * ctx, Gsasl_session * sctx,
+			  Gsasl_property prop)
+{
+  int res;
+
+  /* Call obsolete callbacks. */
+
+  switch (prop)
     {
-      /*
-       * We ended by seeing the end of the string.  Make sure we
-       * have no partial bytes lying around.
-       */
-      if (state != 0)
-	return (-1);
+    case GSASL_VALIDATE_ANONYMOUS:
+      {
+	Gsasl_server_callback_anonymous cb_anonymous;
+	if (!sctx->anonymous_token)
+	  break;
+	cb_anonymous = gsasl_server_callback_anonymous_get (sctx->ctx);
+	if (!cb_anonymous)
+	  break;
+	res = cb_anonymous (sctx, sctx->anonymous_token);
+	return res;
+	break;
+      }
+
+    case GSASL_VALIDATE_EXTERNAL:
+      {
+	Gsasl_server_callback_external cb_external
+	  = gsasl_server_callback_external_get (sctx->ctx);
+	if (!cb_external)
+	  break;
+	res = cb_external (sctx);
+	return res;
+	break;
+      }
+
+    case GSASL_VALIDATE_SECURID:
+      {
+	Gsasl_server_callback_securid cb_securid
+	  = gsasl_server_callback_securid_get (sctx->ctx);
+#define MAX_SECURID 32 /* See RFC 2808. */
+	char buf[MAX_SECURID + 1];
+	size_t buflen = MAX_SECURID;
+	if (!cb_securid)
+	  break;
+	res = cb_securid (sctx, sctx->authid, sctx->authzid, sctx->passcode,
+			  sctx->pin, buf, &buflen);
+	if (buflen > 0 && buflen < MAX_SECURID)
+	  {
+	    buf[buflen] = '\0';
+	    gsasl_property_set (sctx, GSASL_SUGGESTED_PIN, buf);
+	  }
+	return res;
+	break;
+      }
+
+    case GSASL_VALIDATE_GSSAPI:
+      {
+	Gsasl_server_callback_gssapi cb_gssapi
+	  = gsasl_server_callback_gssapi_get (sctx->ctx);
+	if (!cb_gssapi)
+	  break;
+	res = cb_gssapi (sctx, sctx->gssapi_display_name, sctx->authzid);
+	return res;
+	break;
+      }
+
+    case GSASL_VALIDATE_SIMPLE:
+      {
+	Gsasl_server_callback_validate cb_validate
+	  = gsasl_server_callback_validate_get (sctx->ctx);
+	if (!cb_validate)
+	  break;
+	res = cb_validate (sctx, sctx->authzid, sctx->authid, sctx->password);
+	return res;
+	break;
+      }
+
+    case GSASL_PASSWORD:
+      {
+	Gsasl_server_callback_retrieve cb_retrieve
+	  = gsasl_server_callback_retrieve_get (sctx->ctx);
+	char *buf;
+	size_t buflen = BUFSIZ - 1;
+	if (!cb_retrieve)
+	  break;
+	buf = malloc (BUFSIZ);
+	if (!buf)
+	  return GSASL_MALLOC_ERROR;
+	res = cb_retrieve (sctx, sctx->authid, sctx->authzid,
+			   sctx->hostname, buf, &buflen);
+	if (res == GSASL_OK)
+	  gsasl_property_set_raw (sctx, GSASL_PASSWORD, buf, buflen);
+	/* FIXME else if (res == GSASL_TOO_SMALL_BUFFER)... */
+	free (buf);
+	return res;
+	break;
+      }
+
+#if USE_DIGEST_MD5
+    case GSASL_QOPS:
+      {
+	Gsasl_server_callback_qop cb_qop
+	  = gsasl_server_callback_qop_get (sctx->ctx);
+	Gsasl_qop qops;
+	if (!cb_qop)
+	  break;
+	qops = cb_qop (sctx);
+	if (qops & 0x07)
+	  gsasl_property_set (sctx, GSASL_QOPS, digest_md5_qops2qopstr(qops));
+	return GSASL_OK;
+	break;
+      }
+#endif
+
+    default:
+      break;
     }
 
-  return (tarindex);
+  return GSASL_NO_CALLBACK;
 }
